package main

import (
	"bytes"
	"bufio"
	"strconv"
	"log"
	"strings"
	"sort"
	"fmt"
	"os"
	"time"
	"path/filepath"
	"html/template"
)

type tukey5 struct {
	Lo float64
	Q1 float64
	Q2 float64
	Q3 float64
	Hi float64
}

type key struct {
	name string
	size int
}

var names []string
var haveName = map[string]bool{}
var sizes []int
var benches = map[key]tukey5{}

func main() {
	files, _ := filepath.Glob("bench386.go1386.test.*")
	for _, filename := range files {
		size, _ := strconv.Atoi(strings.TrimPrefix(filename, "bench386.go1386.test."))
		if size == 0 {
			log.Fatal("invalid file name: ", filename)
		}
		sizes = append(sizes, size)
		file, err := os.Open(filename)
		if err != nil {
			log.Fatal(err)
		}
		result := map[string][]float64{}
		scan := bufio.NewScanner(file)
		for scan.Scan() {
			f := strings.Fields(scan.Text())
			if len(f) >= 4 && f[3] == "ns/op" {
				ns, err := strconv.ParseFloat(f[2], 64)
				if err != nil {
					log.Fatal("invalid measurement: %s ns/op", f[2])
				}
				result[f[0]] = append(result[f[0]], ns)
				haveName[f[0]] = true
			}
		}
		for name, times := range result {
			sort.Float64s(times)
			benches[key{name, size}] = tukey5{times[0], times[len(times)/4], times[len(times)/2], times[len(times)*3/4], times[len(times)-1]}
		}
	}
	
	for name := range haveName {
		names = append(names, name)
	}
	sort.Strings(names)
	sort.Ints(sizes)
	
	var data = struct{
		Names []string
		Sizes []int
	}{
		names,
		sizes,
	}
	
	dstring := func(d float64) string {
		dt := time.Duration(d)
		scale := 1*time.Second
		for scale > dt {
			scale /= 10
		}
		scale /= 100
		if scale > 0 {
			dt = (dt + scale/2) / scale * scale
		}
		return dt.String()
	}
	
	funcs := template.FuncMap{
		"data": func(name string, size int) tukey5 {
			base := benches[key{name, sizes[0]}].Q2
			t := benches[key{name, size}]
			return tukey5{t.Lo/base, t.Q1/base, t.Q2/base, t.Q3/base, t.Hi/base}
		},
		"rawdata": func(name string, size int) tukey5 {
			return benches[key{name, size}]
		},
		"dur": func(f1, f2 float64) string {
			return fmt.Sprintf("%+.2f%% (%s)", f1*100 - 100, dstring(f2))
		},
		"div": func(x, y int) int { return x / y },
		"base": func(name string) string {
			base := benches[key{name, sizes[0]}].Q2
			return dstring(base)
		},
	}
	graph := template.Must(template.New("main").Funcs(funcs).Parse(graphTmpl))

	var buf bytes.Buffer
	if err := graph.Execute(&buf, data); err != nil {
		log.Fatal(err)
	}
	os.Stdout.Write(buf.Bytes())
	return

	for _, name := range names {
		for _, size := range sizes {
			base := benches[key{name, sizes[0]}].Q2
			t := benches[key{name, size}]
			fmt.Printf("%s %d %.3f %.3f %.3f %.3f %.3f\n",
				name, size, t.Lo/base, t.Q1/base, t.Q2/base, t.Q3/base, t.Hi/base)
		}
	}
}

var graphTmpl = `
<html>
	<head>
		<script type="text/javascript" src="https://www.google.com/jsapi"></script>
		<script type="text/javascript">
			google.load("visualization", "1", {packages:["corechart"]});
			google.setOnLoadCallback(drawCharts);
			function drawCharts() {
{{range $name := $.Names}}
				(function() {
					var div = document.getElementById('graph_{{$name}}');
					if(div == null) {
						return;
					}
					var data = new google.visualization.DataTable();
					data.addColumn('string', 'Stack Size (kilobytes)');
					{{/* data.addColumn('number', 'Slowest'); */}}
					data.addColumn('number', '25%');
					data.addColumn({type: 'string', role: 'tooltip'});
					data.addColumn('number', 'Median');
					data.addColumn({type: 'string', role: 'tooltip'});
					data.addColumn('number', '75%');
					data.addColumn({type: 'string', role: 'tooltip'});
					{{/* data.addColumn('number', 'Fastest'); */}}
					data.addRows([
{{range $size := $.Sizes}}{{$d := data $name $size}}{{$rd := rawdata $name $size}}
						[{{printf "%v" (div $size 1024)}}, {{/* {{$d.Lo}}, */}} {{$d.Q1}}, {{dur $d.Q1 $rd.Q1}}, {{$d.Q2}}, {{dur $d.Q2 $rd.Q2}}, {{$d.Q3}}, {{dur $d.Q3 $rd.Q3}} {{/* , {{$d.Hi}} */}}],
{{end}}
					]);
					var options = {
						width: 800, height: 300,
						title: '{{$name}} (base = {{base $name}})',
						areaOpacity: 0.0,
						focusTarget: 'category',
						strictFirstColumnType: true,
						hAxis: {minValue: 0, maxValue: 64, viewWindowMode: 'pretty', title: 'stack size (kilobytes)'},
						vAxis: {minValue: 0.9, ticks: [{v: 0.0, f: "0.0"}, {v: 0.1, f: ""}, {v: 0.2, f: "0.2"}, {v: 0.3, f: ""}, {v: 0.4, f: "0.4"}, {v: 0.5, f: ""}, {v: 0.6, f: "0.6"}, {v: 0.7, f: ""}, {v: 0.8, f: "0.8"}, {v: 0.9, f: ""}, {v: 1.0, f: "1.0"}, {v: 1.1, f: ""}]},
						vAxes: {0: {title: 'normalized run time'}},
					};
					var chart = new google.visualization.AreaChart(div);
					chart.draw(data, options);
				})();
{{end}}
			}
		</script>
		<style>
		</style>
	</head>
	<body>
	<h2>Go 1 benchmark speed as a function of stack frame size</h2>
	<p>
	Russ Cox
	<br>
	October 2013
	</p>
	
	<div>
	<p>Benchmarks were run on a 2.27 GHz Intel Xeon E5520 with four cpus of four cores each using tip as of October 1 (664c8763de42). GOOS=linux, GOARCH=386, GOMAXPROCS=1.
	</p>
	<p>The &ldquo;extra stack&rdquo; added to a jumbo frame (the padding added for frame allocated for a single function that needs more than the minimum frame size) varied along
	with the frame size, at frame size / 4, but capped at 8kB (so a 4kB minimum frame has a 1kB additional stack, 8kB has 2kB, ..., 32kB and above have 8kB).</p>
	<p>
	Each benchmark was run 12 times with -test.benchtime=5s. The graph plots the middle three quartiles of the runs (25%, 50%, 75%), normalized to the median run time of the 4kB stack run.
	Hover over the lines to see raw data.
	</div>

{{range $name := $.Names}}
	<div id="graph_{{$name}}"></div>
{{end}}
	</body>
</html>
`
